/*
Copyright (c) 2003 Jasper van de Gronde. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
   3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef BUFFERLIB_BUFFERIO_HPP
#define BUFFERLIB_BUFFERIO_HPP

//--------------------------------------------------------------------------- : Includes

#include "bufferlibtypes.hpp"
#include "buffer.hpp"

//--------------------------------------------------------------------------- : Namespace

namespace bufferlib
{

//--------------------------------------------------------------------------- : Input types

//! Input class (lets you WRITE to the buffer)
template<typename T>
class input_t
{
protected:
    buffer_t<T>* const buffer_;
    count_t position_;

public:
    input_t(buffer_t<T>* const buffer) :
            buffer_(buffer),
            position_(0)
    {
        //		buffer_->add_input(shared_from_this());
    }
    ~input_t()
    {} // I don't have to remove myself from buffer_, as buffer_ keeps a shared_ptr of me.
    virtual count_t write(const T* const buffer, const count_t count)
    {
        return buffer_->write(buffer, count, position_); // Also updates position_!
    }
    inline count_t position() const
    {
        return position_;
    }
};

//! Output class (lets you READ from the buffer)
template<class T>
class output_t
{
protected:
    buffer_t<T>* const buffer_;
    count_t position_;

public:
    output_t(buffer_t<T>* const buffer) :
            buffer_(buffer),
            position_(0)
    {
        //		buffer_->add_output(shared_from_this());
    }
    ~output_t()
    {} // I don't have to remove myself from buffer_, as buffer_ keeps a shared_ptr of me.
    virtual count_t read(T* const buffer, const count_t max_count)
    {
        return buffer_->read(buffer, max_count, position_);
    }
    inline count_t position() const
    {
        return position_;
    }
};

//--------------------------------------------------------------------------- : EOF

} // Namespace bufferlib

#endif //BUFFERLIB_BUFFERIO_HPP
