/*
Copyright (c) 2003 Jasper van de Gronde. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
   3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef BUFFERLIB_RINGBUFFER_HPP
#define BUFFERLIB_RINGBUFFER_HPP

//--------------------------------------------------------------------------- : Includes

#include "buffer.hpp"

//--------------------------------------------------------------------------- : Namespace

namespace bufferlib
{

//--------------------------------------------------------------------------- : ringbuffer_t

template<typename T>
class ringbuffer_t :
            public buffer_t<T>
{
protected:
    T* buffer_;
    count_t buffer_size_;

public:
    count_t lowest_input_position_in_buffer() const
    {
        /*if ( !input_positions_.empty() ) {
        	count_t min=*(input_positions_.begin()) % buffer_size_;
        	for(std::multiset<count_t>::const_iterator it=++(input_positions_.begin()); it!=input_positions_.end(); ++it)
        	{
        		if ( (*it % buffer_size_) < min ) {
        			min = (*it % buffer_size_);
        		}
        	}
        	return min;
        } else {
        	throw std::runtime_error("No input positions!");
        }*/
        return lowest_input_position() % buffer_size_;
    }
    //! Returns the input_position belonging to the lowest_input_position_in_buffer.
    count_t lowest_input_position_in_buffer_ex() const
    {
        /*if ( !input_positions_.empty() ) {
        	count_t min=*(input_positions_.begin());
        	for(std::multiset<count_t>::const_iterator it=++(input_positions_.begin()); it!=input_positions_.end(); ++it)
        	{
        		if ( (*it % buffer_size_) < min % buffer_size_ ) {
        			min = *it;
        		}
        	}
        	return min;
        } else {
        	throw std::runtime_error("No input positions!");
        }*/
        return lowest_input_position();
    }
    count_t lowest_output_position_in_buffer() const
    {
        /*if ( !output_positions_.empty() ) {
        	count_t min=*(output_positions_.begin()) % buffer_size_;
        	for(std::multiset<count_t>::const_iterator it=++(output_positions_.begin()); it!=output_positions_.end(); ++it)
        	{
        		if ( (*it % buffer_size_) < min ) {
        			min = (*it % buffer_size_);
        		}
        	}
        	return min;
        } else {
        	throw std::runtime_error("No output positions!");
        }*/
        return lowest_output_position() % buffer_size_;
    }
    //! Returns the output_position belonging to the lowest_output_position_in_buffer.
    count_t lowest_output_position_in_buffer_ex() const
    {
        /*if ( !output_positions_.empty() ) {
        	count_t min=*(output_positions_.begin());
        	for(std::multiset<count_t>::const_iterator it=++(output_positions_.begin()); it!=output_positions_.end(); ++it)
        	{
        		if ( (*it % buffer_size_) < min % buffer_size_ ) {
        			min = *it;
        		}
        	}
        	return min;
        } else {
        	throw std::runtime_error("No output positions!");
        }*/
        return lowest_output_position();
    }

public:
    virtual count_t write(const T* const buffer, const count_t param_count, count_t& position)
    {
        const count_t count = std::min(can_write_from(position),param_count);
        const count_t position_in_buffer = position % buffer_size_;
        const count_t spare_space = buffer_size_ - position_in_buffer;
        if ( count <= spare_space )
        {
            memcpy(buffer_ + position_in_buffer, buffer, count * sizeof(T));
        }
        else
        {
            memcpy(buffer_ + position_in_buffer, buffer, spare_space * sizeof(T));
            memcpy(buffer_, buffer + (spare_space * sizeof(T)), (count - spare_space) * sizeof(T));
        }
        update_input_position(position,count); // Updates position if it needs to.
        return count;
    }

    virtual count_t read(T* const buffer, const count_t param_max_count, count_t& position)
    {
        const count_t count = std::min(can_read_from(position),param_max_count);
        const count_t position_in_buffer = position % buffer_size_;
        const count_t spare_space = buffer_size_ - position_in_buffer;
        if ( count <= spare_space )
        {
            memcpy(buffer, buffer_ + position_in_buffer, count * sizeof(T));
        }
        else
        {
            memcpy(buffer, buffer_ + position_in_buffer, spare_space * sizeof(T));
            memcpy(buffer + (spare_space * sizeof(T)), buffer_, (count - spare_space) * sizeof(T));
        }
        update_output_position(position,count); // Updates position if it needs to.
        return count;
    }

    virtual count_t can_write_from(const count_t position) const
    {
        const count_t position_in_buffer = position % buffer_size_;
        const count_t count = (	lowest_output_position_in_buffer() > position_in_buffer ||
                                (lowest_output_position_in_buffer() == position_in_buffer && lowest_output_position_in_buffer_ex() < position) ) ?
                              (lowest_output_position_in_buffer() - position_in_buffer) :
                              (buffer_size_ - position_in_buffer + lowest_output_position_in_buffer());
        return count;
    }
    virtual count_t can_read_from(const count_t position) const
    {
        const count_t position_in_buffer = position % buffer_size_;
        const count_t count = (	lowest_input_position_in_buffer() > position_in_buffer ||
                                lowest_input_position_in_buffer_ex() == position ) ?
                              (lowest_input_position_in_buffer() - position_in_buffer) :
                              (buffer_size_ - position_in_buffer + lowest_input_position_in_buffer());
        return count;
    }

    ringbuffer_t(count_t buffer_size)
    {
        buffer_ = new T[buffer_size];
        buffer_size_ = buffer_size;
    }
    virtual ~ringbuffer_t()
    {
        delete[] buffer_;
    }
};

//--------------------------------------------------------------------------- : EOF

} // Namespace bufferlib

#endif //BUFFERLIB_RINGBUFFER_HPP
